import cv2 as cv
import os

from photo4d.Process import pictures_array_from_file
from photo4d.Image_utils import load_date
import photo4d.XML_utils as uxml
from photo4d.Utils import exec_mm3d

from shutil import rmtree
import pandas as pd
import numpy as np


"""
Compute GCP positions using a MicMac command : Tapioca 
"""


def cut_image(image_path, pos, kernel_size, output_name="", output_folder="/"):
    """
    extract a portion of an image, centered on pos, of a given size (works for .JPG only)
    :param image_path: path of the input image
    :param pos: tuple, position for the center of output image (xpos, ypos)
    :param kernel_size: tuple, size of the output image, in pixels
    :param output_name: full name for output file (ex: "output.JPG")
    :param output_folder: folder for saving output, must already exist
    """
    if output_name == "": output_name = image_path.split('/')[-1].split(".")[-2] + "_cut" + ".JPG"
    if not os.path.exists(output_folder): os.makedirs(output_folder)
    img = cv.imread(image_path)
    pos = int(pos[0]), int(pos[1])
    ysize, xsize = img.shape[0], img.shape[1]
    if (0 <= pos[0] <= ysize) and (0 <= pos[1] <= xsize):
        xmin, xmax = pos[1] - kernel_size[1] // 2, (pos[1] + kernel_size[1] // 2)
        ymin, ymax = pos[0] - kernel_size[0] // 2, pos[0] + kernel_size[0] // 2

        output = img[ymin:ymax, xmin:xmax]

        cv.imwrite(output_folder + output_name, output)
    else:
        print("\033[0;31Position {} not in the picture {}, with size {}\image ignored\033[0m".format(pos, image_path,
                                                                                                     img.shape))


def detect_from_s2d_xml(s2d_xml_path, folder_list, pictures_array, samples_folder_list=None,
    kernel_size=(200, 200), display_micmac=False):
    """
    write extracts files from an xml with image position, and launch detection of the points for all files in folder
    :param s2d_xml_path: file created by the function SaisieAppuisInitQT, in MicMac
    :param folder_list: list of folders containing pictures. One folder is for one camera
    :param samples_folder_list: list of folder where to save samples, in the same order as folder_list
            by default, create "Samples/" folder in each camera folder
    :param pictures_array: array containing names of pictures to process
            each row is considered as a set, and pictures names must be in the same order as secondary_folder_list
            the first item of a row is a boolean, indicating if the set is valid or not
    :param kernel_size: size of the portion of picture to cut for detection (in pixels)
    :param display_micmac: to activate or stop printing MicMac log
    :return:
    Data Frame where each row store initial and detected image position of the tie points :
    [image name, gcp name, TP coord X in image_ini, TP Y ini,index of folder/camera, date,
     TP coord X in image detect, TP Y detect, coord X GCP ini, Y GCP ini]
    """
    # do some checks, and set default values for sample_folder_list
    if s2d_xml_path[-4:] != ".xml": raise IOError("The parameter S2D_xml_path must be an .xml file")
    nb_folders = len(folder_list)
    for i in range(nb_folders):
        folder = folder_list[i]
        if not os.path.exists(folder):
            raise IOError("Invalid path " + folder + " in folder_list")
        if folder[-1] != "/": folder_list[i] += "/"

    if samples_folder_list is None:
        samples_folder_list = []
        for i in range(nb_folders):
            samples_folder_list.append(folder_list[i] + "Samples/")
    elif nb_folders != len(samples_folder_list):
        print("WARNING the parameter samples_folder_list must have the same number of folders as folder_list")
        exit(1)
    for i in range(nb_folders):
        samples_folder = samples_folder_list[i]
        if not os.path.exists(samples_folder):
            try:
                os.makedirs(samples_folder)
            except IOError:
                print("WARNING Invalid path " + samples_folder + " in samples_folder_list")
                exit(1)
        if samples_folder[-1] != "/": samples_folder_list[i] += "/"

    # ==================================================================================================================
    # collect data from xml
    dict_img = uxml.read_S2D_xmlfile(s2d_xml_path)
    panda_result = []  # store all the results
    # iterate over pictures
    for image in dict_img.keys():

        dict_measures = dict_img[image]

        # try to found the image in all folders (assuming all pictures have different names)
        # if the image is found, do the detection in this folder
        found = False
        i = 0
        while not found and i < nb_folders:
            if image in os.listdir(folder_list[i]):
                found = True
                # if the image is found, launch the detection
                #  ==============================================
                print("\nDetection launched for picture {} as reference...".format(image))
                for gcp in dict_measures.keys():
                    print("\n  Detection of point {} in folder  {}/{}".format(gcp, i + 1, nb_folders))
                    pos_ini = dict_measures[gcp]
                    date = load_date(folder_list[i] + image)
                    # add a line for the master image, with the gcp position, because micmac won't launch
                    # the detection on this one, but the point coordinates are still useful
                    panda_result.append(
                        [image, gcp, kernel_size[0]/2, kernel_size[1]/2, i, date,
                         kernel_size[0] / 2, kernel_size[1] / 2, pos_ini[0], pos_ini[1], image])

                    # creation of extract for each picture of the folder, around the point initial position
                    print("  Create extract for detection :\n")
                    images_list = []  # store pictures to use for detection
                    for line in pictures_array:
                        if line[0]:
                            print("  - " + line[i + 1] + "...  ", end="")

                            cut_image(folder_list[i] + line[i + 1], (pos_ini[1], pos_ini[0]), kernel_size=kernel_size,
                                      output_folder=samples_folder_list[i] + gcp + "/", output_name=line[i + 1])
                            images_list.append(line[i + 1])
                            print("done")

                    # launch Tapioca on the selected files
                    #  ==============================================
                    # create file telling MicMac which files to process
                    uxml.write_couples_file(samples_folder_list[i] + gcp + "/" + "couples.xml", image, images_list)
                    os.chdir(samples_folder_list[i] + gcp + "/")
                    print("\n  Launching MicMac...")
                    command = "mm3d Tapioca File couples.xml -1 ExpTxt=1"
                    success, error = exec_mm3d(command, display_micmac)

                    # read results and append it to result
                    #  ==============================================
                    print(success)
                    if success == 0:
                        print("  Tapioca executed with success, reading results")
                        # read output txt files
                        for picture_recap in os.listdir("Homol/Pastis" + image):
                            if picture_recap.split(".")[-1] == "txt":
                                tie_points = uxml.get_tiepoints_from_txt("Homol/Pastis" + image + "/" + picture_recap)
                                date = load_date(folder_list[i] + picture_recap[:-4])

                                for tie_point in tie_points:
                                    # append each tie point coordinates to the result
                                    # [image name, gcp name, TP coord X in image_ini, TP Y ini,index of folder/camera,
                                    #  date, TP coord X in image detect, TP Y detect, coord X GCP ini, Y GCP ini]
                                    panda_result.append(
                                        [".".join(picture_recap.split(".")[:-1]), gcp, tie_point[0], tie_point[1], i,
                                         date,
                                         tie_point[2], tie_point[3], pos_ini[0], pos_ini[1], image])
                        try:
                            rmtree("Pastis")
                            rmtree("Tmp-MM-Dir")
                        except PermissionError:
                            print("  couldn't erase temporary files due to permission error")
                    else:
                        print("  WARNING Fail in Tapioca :  " + str(error))

            else:
                i += 1
        if not found:
            print("\033[0;31Picture {} cannot be find in folder_list\033[0m".format(image))

    return pd.DataFrame(panda_result,
                        columns=['Image', 'GCP', 'Xini', 'Yini', 'folder_index', 'date', 'Xdetect', 'Ydetect',
                                 'Xgcp_ini', 'Ygcp_ini', 'Image_ini'])


def extract_values(df, magnitude_max=50, nb_values=5, max_dist=50, kernel_size=(200, 200), method="Median"):
    """
    extract detected positions from the DataFrame containing tie points coordinates
    feel free to add new methods
    :param df: DataFrame like the one from detect_from_s2d()
    :param magnitude_max: max value in pixels for the magnitude of the vector (from ini to detect)
    :param nb_values: max values to be used for the method
            the values used are the closest from the GCP initial position
    :param max_dist: max value in pixel for the distance from the GCP to the vector origin
    :param kernel_size: size of the extracts used for detection (to determine coordinates of gcp in the extracts)
    :param method: method to use for computing positions
    :return: tuple with 2 elements:
    - a dictionary containing the computed position of GCPs in each picture, readable for the others functions,
    indexed first by picture names and then by GCP names
    - a panda DataFrame containing the computed position of GCPs in each picture
    columns :
    ['Image', 'GCP', 'Xpos', 'Ypos', 'nb_tiepoints', 'date','nb_close_tiepoints']
    """

    # compute new positions of GCP according to the shift of each tie point
    df['Xshift'] = df.Xgcp_ini + df.Xdetect - df.Xini
    df['Yshift'] = df.Ygcp_ini + df.Ydetect - df.Yini

    # compute vector module
    df['magnitude'] = np.sqrt((df.Xini - df.Xdetect) ** 2 + (df.Yini - df.Ydetect) ** 2)

    # compute vector direction
    df['direction'] = np.arctan2((df.Xini - df.Xdetect), (df.Yini - df.Ydetect)) * 180 / np.pi + 180

    # compute from gcp and tie point in the initial image (gcp is in the center of the extracts)
    pos_center = kernel_size[0] / 2, kernel_size[1] / 2
    df['dist'] = np.sqrt((df.Xini - pos_center[0]) ** 2 + (df.Yini - pos_center[1]) ** 2)

    # filter outliers having a incoherent magnitude
    df_filtered = df.loc[df.magnitude <= magnitude_max]

    dic_image_gcp = {}
    result = []
    # iterate over images
    for image, group in df_filtered.groupby(['Image']):
        dic_gcp = {}
        for gcp, group_gcp in group.groupby(['GCP']):
            nb_tiepoints = group_gcp.shape[0]
            group_gcp_filtered = group_gcp.loc[group_gcp.dist <= max_dist]
            nb_close_tiepoints = group_gcp_filtered.shape[0]
            group2 = group_gcp_filtered.nsmallest(nb_values, 'dist')
            if group_gcp_filtered.shape[0] != 0: # if there is no values left in DataFrame, the point is ignored
                if method == "Median":
                    measure = group2.Xshift.median(), group2.Yshift.median()
                elif method == "Mean":
                    measure = group2.Xshift.mean(), group2.Yshift.mean()
                elif method == 'Min':
                    measure = group2.Xshift.min(), group2.Yshift.min()
                else:
                    print('Method must be one of these values:\n"Median"\n"Min"\n"Mean"')
                    exit(1)
                date = group2.date.min()
                dic_gcp[gcp] = measure


                result.append([image, gcp, measure[0], measure[1], nb_tiepoints, date, nb_close_tiepoints])
        if dic_gcp != {}: dic_image_gcp[image] = dic_gcp


    return dic_image_gcp, pd.DataFrame(result, columns=['Image', 'GCP', 'Xpos', 'Ypos', 'nb_tiepoints', 'date',
                                                        'nb_close_tiepoints'])


if __name__ == "__main__":
    df = detect_from_s2d_xml(
        "C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeurnature/Pictures/Mini_projet/GCP/GCPs_pick-S2D.xml",
        ["C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeurnature/Pictures/Mini_projet/Images/Cam_east",
         "C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeurnature/Pictures/Mini_projet/Images/Cam_mid",
         "C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeurnature/Pictures/Mini_projet/Images/Cam_west"],
        pictures_array=pictures_array_from_file(
            "C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeurnature/Pictures/Mini_projet/set_definition.txt"),
        display_micmac=False
     )
    print(df)
    df.to_csv("C:/Users/Alexis/Documents/Travail/Stage_Oslo/photo4D/python_script/Stats/test_beau.csv", sep=",")
    # df = pd.read_csv("C:/Users/Alexis/Documents/Travail/Stage_Oslo/photo4D/python_script/Stats/test_sift_camtot_new_gcp.csv")
    # result = extract_values(df, threshold=50, nb_values=5, max_dist=200, method="Median")
    # print(result[0])



# coding : uft8
import numpy as np
import os, time
from shutil import copyfile, rmtree, copytree
import photo4d.Image_utils as iu
import photo4d.XML_utils



from photo4d.Image_utils import load_date, load_lum, blurr, lum, process_clahe
from photo4d.Utils import exec_mm3d, pictures_array_from_file


def sort_pictures(folder_path_list, output_path, ext="jpg", time_interval=600):
    """
    Regroup pictures from different folders if they are taken within timeInterval seconds of interval.
    Result is stored in an array/file,
    :param folder_path_list: list of the path to folders containing pictures, each folder is for one camera
    :param output_path: path of the output .txt file containing picture sets
    :param ext: extension of the pictures
    :param time_interval: interval in seconds, with corresponds to the maximum time elapsed between the shooting of pics
    :return: array with the sorted pictures. For each set a boolean is added, always True, but can be modified later
    """
    print("\n Collecting files\n..........................................")
    # create a list containing image names and dates for each folder
    list = []
    for folder_path in folder_path_list:
        image_date_list = []
        flist = os.listdir(folder_path)
        for filename in flist:
            try:
                if filename.split(".")[-1].lower() == ext.lower():
                    image_date_list.append((filename, load_date(os.path.join(folder_path,filename))))
            except IndexError:
                pass
        list.append(image_date_list)
    if len(list) < 1:
        print("WARNING not enough folder\nTwo or more folders are needed to sort files")
        return None
    elif [] in list:
        print("WARNING No image found in One or many folder(s)")
        return None

    sorted_pictures = []
    print(" Checking dates\n..........................................")
    with open(output_path, 'w') as f:
        f.write("# Pictures taken within {} of interval\n".format(time_interval))

    good, bad = 0, 0  # counters for correct and wrong sets
    # loop on the image of the first folder
    for image_ref in list[0]:
        date_ref = image_ref[1]
        pic_group = np.empty(len(list) + 1, dtype=object)
        pic_group[0] = False  # the pic_group[0] is a boolean, True if all a picture is found in every folder
        pic_group[1] = image_ref[0]
        # for_file = [image_ref[0]]  # list of the images taken within the interval

        # check for pictures taken whithin the interval
        for j in range(1, len(list)):
            folder = list[j]  # list of the (filename,date) of one folder
            i, found = 0, False
            while not found and i < len(folder):
                date_var = folder[i][1]
                diff = abs(date_ref - date_var)
                if diff.days * 86400 + diff.seconds < time_interval:  # if the two pictures are taken within 10 minutes
                    found = True
                    pic_group[j + 1] = folder[i][0]
                i += 1

        if None not in pic_group:
            good += 1
            pic_group[0] = True
            print(" Pictures found in every folder corresponding to the timeInterval of " + pic_group[1] + "\n")
            sorted_pictures.append(pic_group)
            with open(output_path, 'a') as f:
                f.write(str(pic_group[0]) + "," + ",".join(pic_group[1:]) + "\n")
        else:
            bad += 1
            print(" Missing picture(s) corresponding to the timeInterval of " + pic_group[1] + "\n")

    end_str = "# {} good set of pictures found, {} uncomplete sets, on a total of {} sets".format(good, bad, good + bad)
    print(end_str)
    with open(output_path, 'a') as f:
        f.write(end_str)
    return np.array(sorted_pictures)


def check_pictures(folder_list, output_path, pictures_array, lum_inf, blur_inf):
    """
    This function is supposed to be called after sort_pictures, as it uses the kind of array created by sort_pictures,
    which could be either collected from the return value of the function, or the file "linked_files.txt" created in
    the main folder
    It will filter pictures in which brightness is inferior to lum_inf and the "blur" (variance of Laplacian) is
    inferior to blur_min
    :param folder_list:
    :param output_path:
    :param pictures_array:
    :param lum_inf:
    :param blur_min:
    :return: same array, but some booleans will be set to False
    """
    print("\n Checking pictures\n..........................................")

    with open(output_path, 'w') as f:
        f.write(
            "# Pictures filtered with a minimum value of {} for brightness, {} for the variance of Laplacian\n".format(
                lum_inf, blur_inf))

    good, bad = 0, 0
    I, J = pictures_array.shape
    for i in range(I):
        if pictures_array[i, 0]:
            min_lum = 9999
            min_blur = 9999
            for j in range(1, J):
                path = os.path.join(folder_list[j - 1],pictures_array[i, j])
                lum = load_lum(path)

                if lum < min_lum:
                    min_lum = lum
                blur = blurr(path, 3)
                if blur < min_blur:
                    min_blur = blur

            if min_lum < lum_inf or min_blur < blur_inf:
                pictures_array[i, 0] = False
                bad += 1
            else:
                good += 1

    with open(output_path, 'a') as f:
        for line in pictures_array:
            f.write(str(line[0]) + "," + ",".join(line[1:]) + "\n")
        end_line = " {} good set of pictures found, {} rejected sets, on a total of {} sets".format(good, bad, good + bad)
        f.write("#" + end_line)
        print(end_line)
    return pictures_array


def copy_and_process(filepath_list, output_folder, tmp_folder_name="TMP_MicMac", ply_name="", clahe=False, tileGridSize_clahe=8, resol=-1,
    distortion_model="RadialStd", incal=None, inori=None, abs_coord_gcp=None,
    img_coord_gcp=None, pictures_ini=None, re_estimate=False,
    master_img=None, masqpath_list=None, DefCor=0.0, shift=None, delete_temp=True,
    display_micmac=True, GNSS_PRECISION=0.2, GCP_POINTING_PRECISION=5):
    """
    copy needed files and process micmac for a set of given pictures (filepath_list)
    It is advised to give only absolute path in parameters


    :param filepath_list: list of absolute path of pictures to process
    :param output_folder: directory for saving results and process pictures
    :param tmp_folder_name: name of the temporary folder where processing pictures
            date is automatically added to the name
    :param ply_name: first part of the name of the output point-cloud, date and extension will be added automatically
    :param clahe: if True, apply a "contrast limited adaptive histogram equalization" on the pictures before processing

    MicMac parameters: (see more documentation on official github and wiki (hopefully))
    :param resol: resolution for the research of TiePoints todo demander à luc pour les racourcis
            -1 stands for full resolution
    :param distortion_model: distortion model used in Tapas
            can be "Fraser", "FraserBasic", "RadialStd", "RadialBasic" and more
    :param incal: path to initial calibration folder (from MicMac), None if no initial calibration used
    :param inori: path to initial orientation folder (from MicMac), None if no initial orientation used
    :param pictures_ini: pictures of initial calibration, in the same order as secondary folder list
    :param re_estimate: -only if there is an initial orientation- if True, re estimate the orientation
    :param abs_coord_gcp: xml file containing absolute coordinates of GCPs, file from the MicMac command GCPConvert
    :param img_coord_gcp: xml file containing image coordinates of the GCPs, result one of the SaisieAppuis command in MicMac

    :param master_img:
    :param masqpath_list: list of mask, each mask file in this list correspond to the picture of the same indice in filepath list
            list of path to .tif files from MicMac
    :param DefCor:
    :param shift: shift for saving ply (if numbers are too big for 32 bit ply) [shiftE, shiftN, shiftZ]
    :param delete_temp: if False the temporary folder will not be deleted, but beware, MicMac files are quite heavy
    :param display_micmac: if False MicMac log will be hidden, may be useful as this may ba a bit messy sometimes
    :return:
    """
    # ==================================================================================================================
    # checking path and parameters :

    # check output folder
    if not os.path.exists(output_folder): os.makedirs(output_folder)

    # check processing folder, name it according to the date of the first picture
    date = load_date(filepath_list[0])
    if date is not None:
        date_str = date.strftime("%Y_%m_%d-%H_%M")
        folder_path = os.path.join (output_folder,tmp_folder_name + "_" + date_str)
    else:
        folder_path = os.path.join(output_folder,tmp_folder_name + filepath_list[0].split('/')[-1])
        date_str = filepath_list[0].split('/')[-1]  # if we can't find a date, put the name of the first picture...
    if not os.path.exists(folder_path): os.mkdir(folder_path)

    # check masks
    if masqpath_list is not None:
        if type(masqpath_list) != list or len(masqpath_list) != len(filepath_list):
            print("Invalid value {} for parameter masqpath_list".format(masqpath_list))
            print("Mask ignored")
            masqpath_list = None

    # ==================================================================================================================
    # copy pictures into the temporary folder
    pictures_pattern = "("  # pattern of pictures telling MicMac which one to process
    for i in range(len(filepath_list)):

        filepath = filepath_list[i]
        filename = os.path.basename(filepath)
        pictures_pattern += filename + "|"

        if not clahe:
            copyfile(filepath, os.path.join(folder_path, filename))

        else:
            # apply the CLAHE method, used for the orientation
            process_clahe(filepath,
                tileGridSize_clahe, out_path=os.path.join(folder_path, filename), grey=True)

        # copy mask corresponding to the picture
        if masqpath_list is not None:
            # copy the mask and gave it the default name of a mask created by MicMac
            copyfile(masqpath_list[i], os.path.join(folder_path, '.'.join(filename.split('.')[:-1]) + "_Masq.tif"))
            # MicMac needs a xml file with the tiff file, if it doesn't exist, it is created
            if os.path.exists('.'.join(masqpath_list[i].split('.')[:-1]) + '.xml'):
                copyfile('.'.join(masqpath_list[i].split('.')[:-1]) + '.xml',
                         os.path.join(folder_path, '.'.join(filename.split('.')[:-1]) + "_Masq.xml"))
            else:
                XML_utils.write_masq_xml(masqpath_list[i], folder_path + '.'.join(filename.split('.')[:-1]) + "_Masq.xml")

    pictures_pattern = pictures_pattern[:-1] + ")"

    # launching MicMac reconstruction
    # ==================================================================================================================

    # detection of Tie points
    # ===================================================
    os.chdir(folder_path)
    print(folder_path)
    # first detection of Tie points with the option ExpTxt=1 to make a report with txt
    command = 'mm3d Tapioca All {} {} ExpTxt=1'.format(pictures_pattern, resol)
    print("\033[0;33" + command + "\033[0m")
    exec_mm3d(command, display_micmac)
    # second detection of Tie points at a lower resolution, without ExpTxt=1, just to avoid a stupid MicMac bug in C3DC
    # MicMac will just convert txt to binary
    command = 'mm3d Tapioca All {} {}'.format(pictures_pattern, resol)
    print("\033[0;33" + command + "\033[0m")
    exec_mm3d(command, display_micmac)

    # Orientation of cameras
    # ===================================================

    # list picture names, used to transform orientation and calibration
    final_pictures = []
    for filepath in filepath_list:
        final_pictures.append(filepath.split("/")[-1])

    ori = distortion_model  # name of orientation, it is the distortion model by default
    # if there is an initial orientation
    print("ICIIII", inori)
    if inori is not None:

        print(os.path.basename(inori))
        # copy the initial orientation file
        new_Ori_path = os.path.join(folder_path, os.path.basename(inori))
        print("OIRIIIII", new_Ori_path)
        if not os.path.exists(new_Ori_path):
            print("COPYTREE",inori,new_Ori_path)
            copytree(inori, new_Ori_path)  # We assume that the path is correctly written ( "/Root/folder/Ori-Truc/")
        # swap pictures names to mock MicMac computed orientation files
        print(pictures_ini)
        print(final_pictures)
        XML_utils.change_Ori(pictures_ini, final_pictures, new_Ori_path)

        # if re_estimate, MicMac recompute the orientation
        if re_estimate:
            ori += "_R"
            command = 'mm3d Tapas {} {} InOri={} Out={}'.format(distortion_model, pictures_pattern,
                                                                os.path.basename(inori), ori)
            print("\033[0;33" + command + "\033[0m")
            success, error = exec_mm3d(command, display_micmac)
        else:
            ori = os.path.basename(inori)  # name of the new ori is the same as the initial one
            success, error = 0, None
    # if there is no initial orientation but an initial calibration
    # it assumed that the calibration is the same for every picture
    elif incal is not None:
        if incal[-1] != "/": incal += "/"
        new_Cal_path = folder_path + incal.split('/')[-2] + "/"
        if not os.path.exists(new_Cal_path):
            copytree(incal, new_Cal_path)  # We assume that the path is correctly written ( "/fgg/Ori-Truc/")
        command = 'mm3d Tapas {} {} InCal={}'.format(distortion_model, pictures_pattern,
                                                     incal.split('/')[-2])
        print("\033[0;33" + command + "\033[0m")
        success, error = exec_mm3d(command, display_micmac)

    # if no initial parameters, classic orientation computation
    else:
        command = 'mm3d Tapas {} {}'.format(distortion_model, pictures_pattern)
        print("\033[0;33" + command + "\033[0m")
        success, error = exec_mm3d(command, display_micmac)

    # if GCP are provided for Bascule (Absolute orientation)
    if abs_coord_gcp is not None and img_coord_gcp is not None:
        # copy GCP absolute coordinates
        gcp_name = os.path.basename(abs_coord_gcp)
        if abs_coord_gcp != os.path.join(folder_path, gcp_name):
            copyfile(abs_coord_gcp, os.path.join(folder_path, gcp_name))
        # create S2D xml file with images positions of gcp
        XML_utils.write_S2D_xmlfile(img_coord_gcp,os.path.join(folder_path, "GCP-S2D.xml"))

        command = 'mm3d GCPBascule {} {} Bascule_ini {} {}'.format(pictures_pattern,
                                                                   ori,
                                                                   gcp_name,
                                                                   "GCP-S2D.xml")
        print(command)
        exec_mm3d(command, display_micmac)
        
        command = 'mm3d Campari {} Bascule_ini Bascule GCP=[{},{},{},{}] AllFree=1'.format(pictures_pattern, gcp_name,GNSS_PRECISION,"GCP-S2D.xml",GCP_POINTING_PRECISION)
        success, error = exec_mm3d(command, display_micmac)

        ori = 'Bascule'

    # dense correlation
    # ===================================================
    if clahe:
        # clahe images were used for orientation, but we need JPG for dense correlation
        for i in range(len(filepath_list)):
            filepath = filepath_list[i]
            filename = os.path.basename(filepath)
            copyfile(filepath, os.path.join(folder_path, filename))

    ply_name += date_str + ".ply"
    command = 'mm3d Malt GeomImage {} {} Master={} DefCor={} MasqIm=Masq'.format(pictures_pattern, ori,
                                                                                 master_img, DefCor)
    print("\033[0;33" + command + "\033[0m")
    success, error = exec_mm3d(command, display_micmac)

    # creation of the final point cloud file
    # ===================================================

    # create the point cloud
    if shift is None:
        command = 'mm3d Nuage2Ply MM-Malt-Img-{}/NuageImProf_STD-MALT_Etape_8.xml Attr={} Out={}'.format(
            '.'.join(master_img.split('.')[:-1]), master_img, ply_name)
    else:
        command = 'mm3d Nuage2Ply MM-Malt-Img-{}/NuageImProf_STD-MALT_Etape_8.xml Attr={} Out={} Offs={}'.format(
            '.'.join(master_img.split('.')[:-1]), master_img, ply_name, str(shift).replace(" ", ""))

    print("\033[0;33" + command + "\033[0m")
    success, error = exec_mm3d(command, display_micmac)

    # ==================================================================================================================
    # Make stats about the reconstruction
    with open(date_str + '_recap.txt', 'w') as recap:
        recap.write("Summary of MicMac 3D reconstruction for " + date_str + "\n")
        string = "\nNumber of Tie Points : "
        df, tot = XML_utils.count_tiepoints_from_txt(folder_path)
        string += str(tot) + "\n\nFiles used :\n"
        for i in df.index.values:
            string += "  - {} taken on {} : {} tie points\n".format(i, load_date(folder_path + i), df.loc())
        recap.write(string)

        if os.path.exists(os.path.join(folder_path,ply_name)):
            copyfile(os.path.join(folder_path,ply_name), os.path.join(output_folder,ply_name))
            recap.write("\nStatus  : Success\n")
        else:
            recap.write("\nStatus  : Failure\n")

        if inori is not None:
            recap.write("\n Initial orientation used  : " + inori + "\n")
        elif incal is not None:
            recap.write("\n Initial calibration used  : " + incal + "\n")

        # write a .txt with residuals
        if ori[:4] == "Ori-":
            ori_path = folder_path + ori + "/"
        else:
            ori_path = folder_path + "Ori-" + ori + "/"
        if os.path.exists(ori_path):  # if Tapas command worked
            dict = XML_utils.extract_res(ori_path)
            recap.write("\nNumber of iterations : {}   Average Residual : {}\n".format(dict['nb_iters'], "%.4f" % dict[
                'AverageResidual']))
            recap.write("\nName           Residual     PercOk        Number of Tie Points\n")
            for i in df.index.values:
                recap.write(
                    "{} : {}       {}       {}\n".format(i, "%.4f" % float(dict[i][0]), "%.4f" % float(dict[i][1]),
                                                         dict[i][2]))
    copyfile(os.path.join(folder_path, date_str + '_recap.txt'), os.path.join(output_folder, date_str + '_recap.txt'))

    # try to delete temporary folder
    if delete_temp:
        try:
            rmtree(folder_path)
        except FileNotFoundError:
            pass
        except PermissionError:
            print("Permission Denied, cannot delete " + folder_path)
        except OSError:
            pass
    else:
        # only delete some heavy micmac files
        try:
            rmtree(folder_path + "Pastis")
            rmtree(folder_path + "Tmp-MM-Dir")
            rmtree(folder_path + "Pyram")
        except PermissionError:
            print("Permission Denied, cannot delete some MicMac files")


def process_from_array(folders_list, pictures_array, output_folder, incal=None, inori=None,
    pictures_ini=None, gcp=None, gcp_S2D=None, clahe=False, tileGridSize_clahe=8, resol=-1,distortion_model="Fraser",
    re_estimate=False, master_folder=0, masq2D=None, DefCor=0.0, shift=None, delete_temp=True,
    display_micmac=True, cond=None, GNSS_PRECISION=0.2, GCP_POINTING_PRECISION=5):
    """
    Do the 3D reconstruction of pictures using MicMac

    Mandatory arguments :
    :param folders_list: list of path to picture folders
            pictures are stored in a folder per camera
    :param pictures_array: array containing names of pictures to process
            each row is considered as a set, and pictures names must be in the same order as secondary_folder_list
    :param output_folder: where output file will be saved as well as where the temporary files will be created
    Reconstruction argument :
    :param resol: resolution for the research of TiePoints
            -1 stands for full resolution
    :param distortion_model: distortion model used in Tapas
            can be "Fraser", "FraserBasic", "RadialStd", "RadialBasic" and more
    :param incal: path to initial calibration folder (from MicMac), None if no initial calibration used
    :param inori: path to initial orientation folder (from MicMac), None if no initial orientation used
    :param pictures_ini: pictures of initial calibration, in the same order as secondary folder list
             None if the pictures are in pictures_array, they will be automatically detected
    :param re_estimate: -only if there is an initial orientation- if True, re estimate the orientation
    :param gcp: .xml file containing coordinates of GCPs, file from the MicMac command "GCPConvert"
    :param gcp_S2D: xml file containing image coordinates of the GCPs, result one of the "SaisieAppuis" command in MicMac

    :param master_folder: list index of the folder where "master images" for Malt reconstruction are stored
            by default it is the first folder
    :param masq2D:
    :param shift: shift for saving ply (if numbers are too big for 32 bit ply) [shiftE, shiftN, shiftZ]
    :param delete_temp: if False the temporary folder will not be deleted, but beware, MicMac files are quite heavy
    :param display_micmac: if False MicMac log will be hidden, may be useful as this may ba a bit messy sometimes
    :return:
    """
    # checking parameters :

    # it is assumed that all pictures have the same extension
    ext = pictures_array[0, 1].split('.')[-1].lower()

    nb_folders = len(folders_list)

    for i in range(nb_folders):
        if folders_list[i][-1] != "/": folders_list[i] += "/"
        if not os.path.exists(folders_list[i]):
            print("WARNING the folder {} in folders_list does not exist".format(folders_list[i]))

    if cond is None:
        def cond(datetime): return True

    if type(master_folder) != int or not (0 <= master_folder < nb_folders):
        print("Invalid value {} for parameter master folder, value set to 0".format(master_folder))
        print("must be one index of the array secondary_folder_list")
        master_folder = 0

    # if pictures_ini is not provided but there is an initial orientation, retrieve it from pictures array
    if inori is not None and pictures_ini is None:
        # find the name of one picture used in inori, using Orientation files
        flist = os.listdir(inori)
        image = ""
        for file in flist:
            if len(file) > 12 and file[:12] == "Orientation-":
                image = file[12:-4]  # todo qu'est ce que c'est moche
                break
        # find the row of this picture in pictures_array and use it as pictures_ini
        nb_sets, nb_pics = pictures_array.shape
        i = 0
        found = False
        while i < nb_sets and not found:
            j = 1
            while j < nb_pics and not found:
                if pictures_array[i, j] == image:
                    found = True
                j += 1
            i += 1
        if i == nb_sets - 1 and not found:
            raise ValueError("Cannot find initial orientation pictures in pictures array\n"
                             "Please fill parameter pictures_ini")
        pictures_ini = pictures_array[i - 1, 1:]

    if gcp is not None and gcp_S2D is not None:
        if type(gcp_S2D) == str:
            if gcp[-4:] != ".xml" or not os.path.exists(gcp_S2D):
                print("WARNING Param gcp must be an xml file from the micmac function GCPConvert")
                exit(1)
            else:
                # retrieve gcp measures for all image in a dictionary
                all_gcp = XML_utils.read_S2D_xmlfile(gcp_S2D)
        elif type(gcp_S2D) == dict:
            # retrieve gcp measures for all image in a dictionary
            all_gcp = gcp_S2D
        else:
            print('ERROR Param gcp_s2d must be either path to the xml or a dictionary')
            exit(1)
        if not os.path.exists(gcp):
            print("WARNING Cannot open gcp file at " + gcp)
            exit(1)

    else:
        all_gcp = None

    if masq2D is not None:
        print("Collecting mask path")
        masqpath_list = ['' for i in range(pictures_array.shape[1] - 1)]
        flist = os.listdir(masq2D)
        for file in flist:
            if file[-9:] == "_Masq.tif":
                I, J = pictures_array.shape
                i = 0
                found = False
                while i < I and not found:
                    j = 1
                    while j < J and not found:

                        if pictures_array[i, j].lower() == file[:-9].lower() + '.' + ext:
                            masqpath_list[j - 1] = os.path.join(masq2D, file)
                            found = True
                        j += 1
                    i += 1
        if '' in masqpath_list:
            print("Cannot find pictures corresponding to mask in pictures array")
            exit(1)
    else:
        masqpath_list = None

    # launch the process
    for line in pictures_array:

        if line[0]:  # line[0] is a boolean, True if the set is correct
            list_path = []
            for i in range(1, len(folders_list) + 1):
                list_path.append(folders_list[i - 1] + str(line[i]))
            # determining which of the pictures is the master image for dense correlation (Malt)
            master_img = line[master_folder + 1]
            date = load_date(list_path[0])

            if cond(date):
                info = "Processing pictures taken on " + str(date) + ":\n  "
                for img in line[1:]:
                    info += img + "  "

                print(info)

                if all_gcp is not None:
                    # create a dictionary with gcp image coordinates of this picture set
                    try:
                        set_gcp = dict([(k, all_gcp.get(k)) for k in line[1:]])
                    except AttributeError:
                        print("Couldn't find gcp measures for this set of pictures in S2D xml\nSet ignored")
                        continue
                else:
                    set_gcp = None

                copy_and_process(
                    list_path,
                    output_folder=output_folder, incal=incal, inori=inori, pictures_ini=pictures_ini, abs_coord_gcp=gcp,
                    img_coord_gcp=set_gcp, re_estimate=re_estimate,
                    master_img=master_img, masqpath_list=masqpath_list,
                    resol=resol, distortion_model=distortion_model,
                    DefCor=DefCor, clahe=clahe, tileGridSize_clahe=tileGridSize_clahe,
                    shift=shift, delete_temp=delete_temp, display_micmac=display_micmac,
                    GNSS_PRECISION=GNSS_PRECISION, GCP_POINTING_PRECISION=GCP_POINTING_PRECISION)


if __name__ == "__main__":
    tic = time.time()
    print("process launched")

    # main_GCP_estim("C:/Users/Alexis/Documents/Travail/Stage_Oslo/Test_scene_blindern/Pictures/ByCam/Cam2/","DSC00918_11h30.JPG",points = 'selection-S2D.xml')
    # sort_pictures(["C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeur nature/Pictures/cam_est/",
    #              "C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeur nature/Pictures/cam_ouest/",
    #              "C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeur nature/Pictures/cam_mid/"])

    array = pictures_array_from_file(
        "C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeurnature/Pictures/linked_files.txt")
    main_folder = "C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeurnature/Pictures/"
    folders = [main_folder + "cam_est/",
               main_folder + "cam_ouest/",
               main_folder + "cam_mid/"]
    output_folder = "C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeurnature/Pictures/TestdelaMortquitue/Results/"
    masq2D = "C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeurnature/Pictures/TestdelaMortquitue/Mask/"
    inori = "C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeurnature/Pictures/TestdelaMortquitue/Ori-Fraser/"
    S2D = "C:/Users/Alexis/Documents/Travail/Stage_Oslo/photo4D/python_script/Stats/All_points-S2D.xml"
    truc = "C:/Users/Alexis/Documents/Travail/Stage_Oslo/Grandeurnature/GCP/Pt_gps_gcp.xml"

    process_from_array(folders, array,
                       output_folder=output_folder, resol=2000, master_folder=2,
                       masq2D=masq2D, DefCor=0.4, inori=inori, re_estimate=True,
                       clahe=True, delete_temp=False, gcp_S2D=S2D, gcp=truc,
                       distortion_model="Fraser", display_micmac=True, GNSS_PRECISION=0.2, GCP_POINTING_PRECISION=5)
    # [410000, 6710000, 0]
    toc = time.time()
    temps = abs(toc - tic)
    print("Executed in {} seconds".format(round(temps, 3)))



# coding: utf-8
"""
Some useful functions
"""
from subprocess import Popen, PIPE, STDOUT
import sys
import numpy as np
from photo4d.Image_utils import load_date



def exec_mm3d(command, display=True):
    """
    launch a MicMac command.
    As MicMac handle errors and wait for user's input, we needed a way to carry on when a command fails
    This function will kill the process if an error happens, allowing further process to be done
    :param command: MicMac command line, string beginning with mm3d
     to see every command allowed, see https://micmac.ensg.eu/index.php/Accueil, or Github
    :param display: display or not the logs from MicMac, boolean
    :return:
    """
    if command[:5] != "mm3d ":
        print("WARNING The command must begin with mm3d")
        exit(1)

    process = Popen(command.split(" "), stdout=PIPE, stderr=STDOUT)
    for line_bin in iter(process.stdout.readline, b''):
        try:
            line = line_bin.decode(sys.stdout.encoding)
            if display:
                sys.stdout.write(line)
            # if waiting for input, which means an error was generated
            if '(press enter)' in line:
                print("Error in MicMac process, abort process")
                process.kill()
                return 1, None
            elif 'Warn tape enter to continue' in line:
                print("Value error in Tapas, abort process")
                process.kill()
                return 1,'Value error in Tapas'
        except UnicodeDecodeError:
            sys.stdout.write('---cannot decode this line---')

    # if no error occurs
    return 0, None


def pictures_array_from_file(filepath):
    """
    Could be way more efficient ! And does not handle any error for the moment
    :param filepath:
    :return:
    """
    print("\nRetrieving data from file " + filepath + "\n.......................................")
    all_lines = []
    with open(filepath, 'r') as file:
        for line in file.readlines():
            if line[0] != "#" :
                list_temp = line.split(',')
                length = len(list_temp)
                array_line = np.empty(length, dtype=object)
                if list_temp[0].rstrip(" ").lower() == "true":
                    array_line[0] = True
                else:
                    array_line[0] = False
                for i in range(1, length):
                    array_line[i] = list_temp[i].rstrip('\n')
                all_lines.append(array_line)
        print("Done")
        return np.array(all_lines)















# coding : utf8
"""

To keep metadata when we transform the picture, we use the module pyxif (MIT License), available at :
                         https://github.com/zenwerk/Pyxif
"""


from datetime import datetime
import sys, pyxif, time, os
import cv2 as cv
import numpy as np




def load_date(filename):
    """
    Load date of the shot, according to te image metadata
    :param filename: name/path of the file
    :return: datetime format
    """
    try:
        zeroth_dict, exif_dict, gps_dict = pyxif.load(filename)
        date,time=exif_dict[pyxif.PhotoGroup.DateTimeOriginal][1].split(" ")
        year, month,day = date.split(":")
        hour,minute,sec = time.split(":")
        dateimage= datetime(int(year), int(month), int(day), int(hour), int(minute) ,int(sec))
        return dateimage
    except KeyError:
        print("WARNING No date for file " + filename)
        return None
    except FileNotFoundError:
        print("WARNING Could not find file " + filename )
        return None


def load_lum(filename):
    """
    Load luminosity of the shot scene, according to te image metadata
    :param filename: name/path of the file
    :return: float, level of brightness

    TODO: Add a method to estimate BrightnessValue of image if the field is not available from picture EXIF.
    """
    try:
        zeroth_dict, exif_dict, gps_dict = pyxif.load(filename)
        num,denom=exif_dict[pyxif.PhotoGroup.BrightnessValue][1]
        brightness=num/denom
        return brightness
    except KeyError:
        print("WARNING No brightness data for file " + filename)
        print("Check if your exif data contains a 'BrightnessValue' tag ")
        return None
    except FileNotFoundError:
        print("WARNING Could not find file " + filename )
        return None



def process_clahe_folder(in_folder, tileGridSize, grey=False, out_folder="", clip_limit=2,new_name_end="_Clahe"):
    """
    Apply CLAHE to all jpeg files if a given folder
    It is not possible to overwrite files, because the initial files are needed to copy-past metadata

    :param in_folder: input folder path
    :param tileGridSize: size of the "blocks" to apply local histogram equalization
    :param grey: if True, the image will be converted to grayscale
    :param clip_limit: contrast limit, used to avoid too much noise
    :param out_folder: output folder path
    :param new_name_end: string put at the end of output files, without the extension
    :return:
    """

    # Process all the jpeg pictures in the following folder
    flist = np.sort(os.listdir(in_folder))

    for f in flist:
        try:
            if f.split(".")[-1].lower() in ["jpg","jpeg"]:
                in_path = in_folder + f
                if out_folder == "":  out_folder = in_folder
                out_path = out_folder + f[:-4] + new_name_end + ".JPG"

                process_clahe(in_path, tileGridSize, grey=grey, out_path=out_path, clip_limit=clip_limit)
        except IndexError:
            pass


def process_clahe(in_path, tileGridSize, grey=False, out_path="", clip_limit=2):
    """
    Appy CLAHE (contrast limited adaptive histogram equalization) method on an image
    for more information about CLAHE, see https://docs.opencv.org/3.1.0/d5/daf/tutorial_py_histogram_equalization.html

    Overwriting image will raise an error, as the initial image is needed to copy-past metadata
    :param in_path: input image
    :param tileGridSize: size of the "blocks" to apply local histogram equalization
    :param grey: if True, the image will be converted to grayscale
    :param out_path: output path, the folders must exists and the image extension must be valid
            by default, output will be saved as input_path/input_name_clahe.JPG
    :param clip_limit: contrast limit, used to avoid too much noise
    """
    if out_path == "":
        out_path = ".".join(inpath.split(".")[:-1]) + "_clahe.JPG"

    # read input
    print("Processing CLAHE method on " + in_path.split("/")[-1])
    img = cv.imread(in_path)

    # convert color to gray
    if grey: img = cv.cvtColor(img, cv.COLOR_BGR2GRAY)

    # apply a median filter before clahe
    img = cv.medianBlur(img, 3)  
    
    # create clahe object
    clahe = cv.createCLAHE(clipLimit=clip_limit, tileGridSize=(tileGridSize, tileGridSize))  # CLAHE

    # apply CLAHE for each image channel, and then recreate the full image (only useful if gray==False)
    channels_ini = cv.split(img)
    channels_final = []
    for channel in channels_ini:
        # Apply CLAHE
        channels_final.append(clahe.apply(channel))
    img_final = cv.merge(channels_final)

    # save image and write metadata from initial file
    cv.imwrite(out_path, img_final)
    pyxif.transplant(in_path, out_path)

def lum(filename):
    image_bgr = cv.imread(filename)
    image_lab = cv.cvtColor(image_bgr, cv.COLOR_BGR2LAB)
    average_lum = cv.mean(cv.split(image_lab)[0])
    return average_lum


def blurr(filename,ksize = 3):
    image_bgr = cv.imread(filename)  # todo la converstion en gray devrait être fait à cette ligne
    # image_gray = cv.cvtColor(image_bgr, cv.COLOR_BGR2GRAY)
    return np.log(cv.Laplacian(image_bgr, cv.CV_64F,ksize=ksize).var())


# coding :utf8

from lxml import etree
import os
import pandas as pd
import numpy as np
import cv2 as cv


# todo finir la traduction
def extract_res(ori_folder_path, xmlfile='Residus.xml', do_txt=False, output_folder_path="",
    output_name="residuals_last_iter.txt",
    sep=',', ):
    """
    Read the residual file from MicMac and create a txt file (with separator, allows .csv) with only residuals of the
    last iteration : mean residual and residual for each picture.

    :param ori_folder_path: folder where is the xml residuals file, from MicMac, it is an orientation folder (name beginning with "Ori-)
    :param xmlfile: name of xml file, always 'Residus.xml' from MicMac
    ======
    these parameter are only usefull if do_txt==True, in this case a txt file will be writen
    :param do_txt: if False, no file will be written
    :param output_folder_path: folder where to save the txt file, by default same as ori_folder_path
    :param output_name: name of the output file, default "residuals_last_iter.txt"
    :param sep: separator of output file, default ','
    ======
    :return: 1 if failed, dictionary of residuals if no errors detected
            the dictionary have nb_inter, AverageResidual and names of the pictures as index
            the element corresponding to one picture is a tuple ('Name', 'Residual', 'PercOk', 'NbPts', 'NbPtsMul')
            (for more information about these values, see MicMac documentation)
    """

    file_content = ""
    dict = {}

    elements = ('Name', 'Residual', 'PercOk', 'NbPts', 'NbPtsMul')
    try:
        # Parsing of xml
        tree = etree.parse(ori_folder_path + xmlfile)

        # Getting number of iterations
        nb_iters = tree.xpath("/XmlSauvExportAperoGlob/Iters/NumIter")[-1].text
        file_content += 'nb_iters' + sep + nb_iters + '\n'
        dict['nb_iters'] = int(nb_iters)

        # Recuperation de la moyenne des residus de la derniere iteration
        av_resid = tree.xpath("/XmlSauvExportAperoGlob/Iters[NumIter={}][NumEtape=3]/\
                                    AverageResidual".format(nb_iters))[0].text
        file_content += 'AverageResidual' + sep + av_resid + '\n'
        dict['AverageResidual'] = float(av_resid)

        # Recuperation des donnees pour chaque image de la derniere iteration
        file_content += ('\nName{}Residual{}PercOk{}NbPts{}NbPtsMul\n'.format(sep, sep, sep, sep))
        for img in tree.xpath("/XmlSauvExportAperoGlob/Iters[NumIter={}]\
                                                                 [NumEtape=3]/OneIm".format(nb_iters)):
            obj = ''
            for e in elements:
                obj += img.find(e).text + sep
            file_content += obj + '\n'
            image_name = obj.split(sep)[0]
            dict[image_name] = obj.split(sep)[1:-1]
    except OSError:
        print("WARNING Can't open the file " + ori_folder_path + xmlfile)
        return 1
    except etree.XMLSyntaxError:
        print("WARNING The xml is not correct")
        return 1

    # write the txt file
    if do_txt:
        if output_folder_path == "":
            output_folder_path = ori_folder_path

        # Creation of the txt file
        try:
            with open(output_folder_path + output_name, "w") as file:
                file.write(file_content)
        except IOError:
            print("Cannot write file")

    return dict


def read_S2D_xmlfile(file_path):
    """
    read a file containing pixel coordinates of points in images from the SaisieAppuis MicMac commands and put data
    into a dictionary

    :param file_path: path to the s2d xml file
    ========
    :return: a dictionary, pictures names are the indexes, and for each picture the element is a dictionary with points measurments
    todo cest vraiment trop mal ecrit !!!!!!
    { picture1 : {point1: (coordX, coordY),
                 {point2: (coordX, coordY)}
      picture2 : {point1: (coordX, coordY),
                 ...}
    ...}
    """
    dic_img_measures = {}
    # read the file
    try:
        with open(file_path, 'r') as xml:
            # Parsing of xml
            tree = etree.parse(file_path)
            # loop on images
            for image in tree.xpath("/SetOfMesureAppuisFlottants/MesureAppuiFlottant1Im/NameIm"):
                dic_measures = {}
                # loop on the points
                for point in tree.xpath(
                        "/SetOfMesureAppuisFlottants/MesureAppuiFlottant1Im[NameIm='{}']/OneMesureAF1I".format(
                            image.text)):
                    point_name = point[0].text.rstrip(" ")
                    measure = point[1].text.split(" ")
                    dic_measures[point_name] = (float(measure[0]), float(measure[1]))

                dic_img_measures[image.text] = dic_measures

    except etree.XMLSyntaxError:
        print("WARNING The xml file is not valid  " + file_path)
        exit(1)
    except FileNotFoundError:
        print("WARNING Cannot find file S2D xml at  " + file_path)
        exit(1)

    return dic_img_measures


# todo faire qqchose de cette fonction qui craint...

def count_tiepoints_from_txt(main_folder_path):
    """
    generate a panda DataFrame with the tie points found by the Tapioca command of MicMac USED WITH ExpTxt=1

    :param main_folder_path: path of the folder where is situated the Homol folder, or path of the Homol folder
    :return:
    - a panda DataFrame, row and column indexes are the name of pictures used, and for cell, the number
    of Tie points found in image row compared to img colum todo rendre ca lisible
            Img1    Img2    Img3
    Img1    0       nb1,2   nb1,3
    Img2    nb2,1     0     nb2,3
    Img3    nb3,1   nb3,1     0
    """

    # path checking
    if main_folder_path[-1] != "/": main_folder_path += "/"
    if main_folder_path.split("/")[-2] != "Homol":
        main_folder_path += "Homol/"

    try:
        folder_list = os.listdir(main_folder_path)

        # collect picture names in Homol directory, each folder is for one picture
        index = []
        for folder in folder_list:
            index.append(folder[6:])  # remove Pastis, the folder name being like PastisNamePicture

        df = pd.DataFrame(np.zeros((len(folder_list), len(folder_list))), index=index, columns=index)

        # count tie points
        s = 0  # total tie points
        for folder in folder_list:
            file_list = os.listdir(main_folder_path + folder)
            for filename in file_list:
                if filename.split('.')[-1] == 'txt':
                    file = open(main_folder_path + folder + "/" + filename, 'r')

                    # basically just counting the number of row in each file
                    i = 0
                    for line in file.readlines():
                        i += 1
                        s += 1
                    df.loc[folder[6:], filename.rstrip('.txt')] = i
        if s == 0:
            print('\033[0;31m WARNING, 0 Tie Points found, please check that ExptTxt=1 in Tapioca \033[0m')
        return df, s
    except IOError:
        print('\033[0;31m' + "Cannot open " + main_folder_path + '\033[0m')


def get_tiepoints_from_txt(path):
    point_list = []
    with open(path) as f:
        for line in f.readlines():
            point_list.append(line.split(" "))
    return np.array(point_list).astype(np.float)


def write_S2D_xmlfile(dico_img_measures, file_name):
    """
    Write  an xml file with 2D mesures of points in different images, in a way that MicMac can read it
    :param dico_img_measures: dictionnary containing 2D measures. Must looks like :
    {NameImage1 (String) : {NamePoint1 (String) : (measureX, measureY) (tuple of float),
                           NamePoint2 (String) :  (measureX, measureY) (tuple of float), todo la doc est a continuer
                           ...},
     NameImage2 (String) : {NamePoint1 (String) : measure (String, 'coordPoint1Image2 coordPoint1Image2'),
                           NamePoint2 (String) : measure (String, 'coordPoint2Image2 coordPoint2Image2'),
                           ...}, ...}
    :param file_name: path or name of the output file
    """
    # Creation of the document root
    measures_set = etree.Element('SetOfMesureAppuisFlottants')

    #  iterate over pictures
    for image, dico_measures in dico_img_measures.items():

        img_meas = etree.SubElement(measures_set, 'MesureAppuiFlottant1Im')
        name_img = etree.SubElement(img_meas, 'NameIm')
        name_img.text = image

        # iterate over measures for each picture
        for point, measure in dico_measures.items():

            pt_mes = etree.SubElement(img_meas, 'OneMesureAF1I')
            etree.SubElement(pt_mes, 'NamePt').text = point
            coord_img_pt = etree.SubElement(pt_mes, 'PtIm')
            coord_img_pt.text = "{} {}".format(measure[0], measure[1])

    # open the file for writing
    try:
        with open(file_name, 'w') as file:
            # Header
            file.write('<?xml version="1.0"?>\n')
            # Writing all the text we created
            file.write(etree.tostring(measures_set, pretty_print=True).decode('utf-8'))
    except IOError:
        print('Error while writing file')
        exit(1)


def write_masq_xml(tif_mask, output=""):
    """
    write  default xml file describing the mask from MicMac
    Even if this file seems useless, MicMac can throw an error without this file associated to the mask (in Malt)

    :param tif_mask: path to the MicMac mask, in .tif format
    :param output: path for output xml file
    """
    # do some checks
    if tif_mask.split('.')[-1] not in ["tif", "tiff"]:
        print("Wrong input path " + tif_mask + "\n Must be a .tif file")
        exit(1)
    if output == "":
        output = '.'.join(tif_mask.split('.')[:-1]) + ".xml"
    elif output.split('.')[-1] != "xml":
        print("Wrong output path " + output + "\n Must be a .xml file")
        exit(1)

    file_ori = etree.Element('FileOriMnt')
    name = etree.SubElement(file_ori, 'NameFileMnt')
    name.text = tif_mask

    nb_pix = etree.SubElement(file_ori, 'NombrePixels')
    shape = cv.imread(
        tif_mask).shape  # todo find a easier way
    nb_pix.text = "{} {}".format(shape[1], shape[0])
    # write some default values
    etree.SubElement(file_ori, 'OriginePlani').text = "0 0"
    etree.SubElement(file_ori, 'ResolutionPlani').text = "1 1"
    etree.SubElement(file_ori, 'OrigineAlti').text = "0"
    etree.SubElement(file_ori, 'ResolutionAlti').text = "1"
    etree.SubElement(file_ori, 'Geometrie').text = "eGeomMNTFaisceauIm1PrCh_Px1D"

    # write the xml file
    try:
        with open(output, 'w') as file:
            file.write('<?xml version="1.0"?>\n')
            file.write(etree.tostring(file_ori, pretty_print=True).decode('utf-8'))
    except IOError:
        print('Error while writing file')
        exit(1)


def change_Ori(initial_pictures, final_pictures, ori_folder_path):
    """
    Changes all the files of an Ori- folder from MicMac, in the way that every reference to initial pictures
    is replaced by reference to final_pictures
    WARNING this will totally modify the folder without backup of the initial one, think about make a copy first
    :param initial_pictures: list of initial pictures to be replaced, in the same order as the final one
    :param final_pictures: list of initial pictures to be replaced, in the same order as the final one
    :param ori_folder_path: path of the orientation folder (name beginning with Ori- )
    """
    # some checks
    print( os.path.basename(ori_folder_path)[:4])
    if os.path.basename(ori_folder_path)[:4] != "Ori-":
        print("Ori path is not valid : {}\nYou need to enter the path to the Ori-folder ".format(ori_folder_path))
        exit(1)
    elif len(initial_pictures) != len(final_pictures):
        print("List of input and output pictures must have the same size")
        exit(1)
    nb_pictures = len(initial_pictures)

    # change orientation files
    for j in range(nb_pictures):
        # rename Orientation files
        if os.path.exists(os.path.join(ori_folder_path, 'Orientation-{}.xml'.format(initial_pictures[j]))):
            os.rename(os.path.join(ori_folder_path, 'Orientation-{}.xml'.format(initial_pictures[j])),
                      os.path.join(ori_folder_path, 'Orientation-{}.xml'.format(final_pictures[j])))

    # write a short summary
    with open(os.path.join(ori_folder_path,"log.txt"), 'w') as log:
        log.write("This orientation was not calculated by MicMac with these pictures\n\n")
        log.write("The names of pictures were just changed \n\n")
        for i in range(nb_pictures):
            log.write("{} was replaced by {}\n".format(initial_pictures[i], final_pictures[i]))


def change_xml(initial_pictures, final_pictures, xml_path):
    """
    Replace all occurrences of initial pictures with final pictures in a xml/txt file
    initial pictures[i] will be replaced by final_pictures[i]

    :param initial_pictures: list of pictures to be replaced
    :param final_pictures: list of replacement pictures, in the same order as initial pictures
    :param xml_path: path to the file to process
    """
    # checking length
    if len(initial_pictures) != len(final_pictures):
        print("List of input and output pictures must have the same size")
        exit(1)
    nb_pictures = len(initial_pictures)

    # Read the xml file
    with open(xml_path, 'r') as file:
        file_data = file.read()
    for i in range(nb_pictures):
        # Replace the target string
        file_data = file_data.replace(initial_pictures[i], final_pictures[i])
    # Write the file out again
    with open(xml_path, 'w') as file:
        file.write(file_data)


def write_couples_file(file_path, master_image, pictures_list):
    """
    write an xml file for micmac command Tapioca, for pictures linked to one image

    :param file_path: path to xml file, if it already exists, it will be replaced
    :param master_image: image to compare with all the others
    :param pictures_list:
    """
    root = etree.Element('SauvegardeNamedRel')
    for img in pictures_list:
        couple = etree.SubElement(root, 'Cple')
        couple.text = str(master_image) + " " + str(img)
    with open(file_path, 'w') as xml:
        xml.write('<?xml version="1.0" encoding="UTF_8"?>\n')
        xml.write(etree.tostring(root, pretty_print=True).decode('utf-8'))


if __name__ == "__main__":
    print(read_S2D_xmlfile(
        "C:/Users/Alexis/Documents/Travail/Stage_Oslo/photo4D/python_script/Stats/all_points-S2D_dist_max.xml"))



''' 
Program XXX Part I



'''

# import all function
import pyxif

# import public library
import os
from os.path import join as opj
import numpy as np, pandas as pd
from typing import Union
from shutil import copyfile, rmtree, copytree
import matplotlib.pyplot as plt

# Import project libary
import photo4d.Process as proc
import photo4d.Utils as utils
import photo4d.Detect_Sift as ds
import photo4d.XML_utils
import photo4d.Image_utils as iu


class Photo4d(object):
    # Class constants
    # folders
    IMAGE_FOLDER = 'Images'
    ORI_FOLDER = "Ori-Ini"
    MASK_FOLDER = 'Masks'
    GCP_FOLDER = 'GCP'
    RESULT_FOLDER = "Results"
    # file names
    GCP_COORD_FILE_INIT = 'GCPs_coordinates.xml'
    GCP_COORD_FILE_FINAL = 'GCPs_pick-S3D.xml'
    DF_DETECT_FILE = 'df_detect.csv'
    SET_FILE = 'set_definition.txt'
    GCP_PRECISION=0.2 # GCP precision in m
    GCP_POINTING_PRECISION=10 # Pointing precision of GCPs in images (pixels)
    GCP_PICK_FILE_INI = 'GCPs_pick_Ini.xml'
    GCP_PICK_FILE_INI_2D = 'GCPs_pick_Ini-S2D.xml'
    GCP_PICK_FILE_BASC = 'GCPs_pick_Basc.xml'
    GCP_PICK_FILE_BASC_2D = 'GCPs_pick_Basc-S2D.xml'
    GCP_PICK_FILE = 'GCPs_pick.xml'
    GCP_DETECT_FILE = 'GCPs_detect-S2D.xml'
    GCP_NAME_FILE = 'GCPs_names.txt'
    shift=[410000, 6710000, 0]
    # Parameters
    distortion_model="Figee"
    
    
    def __init__(self, project_path, ext='jpg'):
        if not os.path.exists(project_path):
            print("ERROR The path " + project_path + " doesn't exists")
            exit(1)
            
        # add main folder
        self.project_path = os.path.abspath(project_path)
        print("Creation of object photo4d on the folder " + self.project_path)
        
        # add camera folders
        if os.path.exists(opj(self.project_path, Photo4d.IMAGE_FOLDER)):
            self.cam_folders = [opj(self.project_path, Photo4d.IMAGE_FOLDER, cam) for cam in
                                os.listdir(opj(self.project_path, Photo4d.IMAGE_FOLDER))]
            self.nb_folders = len(self.cam_folders)
            print("Added {} camera folders : \n  {}".format(self.nb_folders, '\n  '.join(self.cam_folders)))
        else:
            print('You must create a folder "' + Photo4d.IMAGE_FOLDER + '/" containing your camera folders')
            exit(1)
            
        # =========================================================================
        # add picture sets
        picture_set_def = opj(self.project_path, Photo4d.SET_FILE)
        if os.path.exists(picture_set_def):
            self.sorted_pictures = utils.pictures_array_from_file(picture_set_def)
            print("Added picture sets from " + picture_set_def)
        else:
            self.sorted_pictures = None
        # set default selected set to the last one
        self.selected_picture_set = -1
        
        # =========================================================================
        # add initial orientation
        if os.path.exists(opj(self.project_path, Photo4d.ORI_FOLDER)):
            print("Added initial orientation")
            self.in_ori = opj(self.project_path, Photo4d.ORI_FOLDER)
        else:
            self.in_ori = None
            
        # =========================================================================
        # add image masks
        if os.path.exists(opj(self.project_path, Photo4d.MASK_FOLDER)):
            self.masks = opj(self.project_path, Photo4d.MASK_FOLDER)
            print("Masks created from ")  # todo add the set of masks (and ori)
        else:
            self.masks = None
            
        # add GCP initial files
        # =========================================================================
        if os.path.exists(opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.GCP_COORD_FILE_INIT)):
            self.gcp_coord_file = opj(self.project_path,Photo4d.GCP_FOLDER, Photo4d.GCP_COORD_FILE_INIT)
            print("Added gcp coordinates file")
        else:
            self.gcp_coord_file = None
        if os.path.exists(opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.GCP_NAME_FILE)):
            self.gcp_names = opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.GCP_NAME_FILE)
        else:
            self.gcp_names = None
        # =========================================================================
        # add result of GCPs detection
        if os.path.exists(opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.DF_DETECT_FILE)):
            self.df_detect_gcp = pd.read_csv(opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.DF_DETECT_FILE))
            print("Added gcp detection raw results")
        else:
            self.df_detect_gcp = None
        if os.path.exists(opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.GCP_DETECT_FILE)):
            self.dict_image_gcp = XML_utils.read_S2D_xmlfile(
                opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.GCP_DETECT_FILE))
            print("Added gcp detection final results")
        else:
            self.dict_image_gcp = None
            
        # extension of the images
        self.ext = ext
        
        # condition on picture dates, to process only a few sets
        self.cond = None
        
    def __str__(self):
        string = "\n=======================================================================\n" \
                 "Project Photo4d located at " + self.project_path + \
                 "\n======================================================================="
        string += "\n Contains {} camera folders : \n   {}".format(self.nb_folders, '\n   '.join(self.cam_folders))
        if self.sorted_pictures is None:
            string += "\n Pictures unsorted"
        else:
            string += "\n Pictures sorted in {} sets ".format(len(self.sorted_pictures))
            string += "\n The current selected set is {}".format(self.sorted_pictures[self.selected_picture_set][1:])
            
        string += "\n=======================================================================\n"
        if self.in_ori is not None:
            string += " Initial orientation computed"
            string += "\n=======================================================================\n"
            
        if self.masks is not None:
            string += " Masks done "
            string += "\n=======================================================================\n"
            
        if self.gcp_coord_file is not None:
            string += " Absolute coordinates of GCPs are given"
            if self.dict_image_gcp is not None:
                string += "\n GCPs image coordinates are computed "
            string += "\n=======================================================================\n"
            
        return string
    
    def sort_picture(self, time_interval=600):
        self.sorted_pictures = proc.sort_pictures(self.cam_folders, opj(self.project_path, Photo4d.SET_FILE),
                                                  time_interval=time_interval,
                                                  ext=self.ext)
        return self.sorted_pictures

    def show_gcp_detected(self, img_filename):
        '''
        Function to plot detected GCPs by SIFT  (TBF)
        '''
        img = cv2.imread(img_filename)

        # read GCP pix location in Pixel coordinate, and displacement in respect to original image

        for index, point in gcp.iterrows():

            cv2.circle(img,(int(point.xpix), int(point.ypix)),4, (0,255,0))

            # add displacement vector
            # Draw a diagonal blue line with thickness of 5 px
            cv2.line(img,(xini,yini),(xfin,yfin),(0,255,0),line_thickness) 


            font = cv2.FONT_HERSHEY_SIMPLEX
            cv2.putText(img,point.gcp_name,(int(point.xpix)+10, int(point.ypix)+10), font, 2,(0,255,0),3,cv2.LINE_AA)

        # Open image with pyplot
        plt.figure()
        plt.imshow(img)
        plt.show()






    def check_picture(self, luminosity_thresh=1, blur_thresh=6):
        '''
        Function to Check if pictures are not too dark and/or too blurry (e.g. fog)
        '''
        if self.sorted_pictures is None:
            print("ERROR You must launch the sort_pictures() method before check_pictures()")
            exit(1)
        self.sorted_pictures = proc.check_pictures(self.cam_folders, opj(self.project_path, Photo4d.SET_FILE),
                                                   self.sorted_pictures,
                                                   lum_inf=luminosity_thresh,
                                                   blur_inf=blur_thresh)
        return self.sorted_pictures

    def initial_orientation(self, resolution=5000, distortion_mode='Fraser', display=True, clahe=False,
                            tileGridSize_clahe=8):
        '''
        Function to initialize camera orientation of the reference set of images using the Micmac command Tapas
        '''

        # add mkdir, and change dir
        tmp_path = opj(self.project_path, "tmp")
        if not os.path.exists(tmp_path): os.makedirs(tmp_path)
        os.chdir(tmp_path)

        # select the set of good pictures to estimate initial orientation
        selected_line = self.sorted_pictures[self.selected_picture_set]
        file_set = "("
        for i in range(len(self.cam_folders)):
            in_path = opj(self.cam_folders[i], selected_line[i + 1])
            out_path = opj(tmp_path, selected_line[i + 1])
            if clahe:
                iu.process_clahe(in_path, tileGridSize_clahe, out_path=out_path)
            else:
                copyfile(in_path, out_path)
            file_set += selected_line[i + 1] + "|"
        file_set = file_set[:-1] + ")"

        # Execute mm3d command for orientation
        success, error = utils.exec_mm3d("mm3d Tapioca All {} {}".format(file_set, resolution), display=display)
        success, error = utils.exec_mm3d(
            "mm3d Tapas {} {} Out={}".format(distortion_mode, file_set, Photo4d.ORI_FOLDER[4:]), display=display)

        ori_path = opj(self.project_path, Photo4d.ORI_FOLDER)
        if success == 0:
            # copy orientation file
            if os.path.exists(ori_path): rmtree(ori_path)
            copytree(opj(tmp_path, Photo4d.ORI_FOLDER), ori_path)
            self.in_ori = ori_path
        else:
            print("ERROR Orientation failed\nerror : " + str(error))

        os.chdir(self.project_path)
        # todo thisraise a permission error despite this chdir I don't know why
        try:
            rmtree(tmp_path)
        except PermissionError:
            print('WARNING Cannot delete temporary folder due to permission error')
            
    def create_mask(self, del_pictures=True):
        '''
        Note : Only the mask of the central (MASTER) image is necessary
        '''
        
        # add mkdir, and change dir
        mask_path = opj(self.project_path, Photo4d.MASK_FOLDER)
        if not os.path.exists(mask_path): os.makedirs(mask_path)
        # select the set of good pictures to estimate initial orientation
        selected_line = self.sorted_pictures[self.selected_picture_set]
        for i in range(len(self.cam_folders)):
            in_path = opj(self.cam_folders[i], selected_line[i + 1])
            out_path = opj(mask_path, selected_line[i + 1])
            copyfile(in_path, out_path)
            ds.exec_mm3d('mm3d SaisieMasqQT {}'.format(out_path))
            if del_pictures:
                os.remove(out_path)
        self.masks = mask_path
        
    def prepare_gcp_files(self, gcp_coords_file, file_format='N_X_Y_Z', display=True):
        '''
        Function to prepare GCP coordinate from a textfile to Micmac xml format. Make sure your text file format is correct
        '''

        if not os.path.exists(opj(self.project_path, Photo4d.GCP_FOLDER)):
            os.makedirs(opj(self.project_path, Photo4d.GCP_FOLDER))
            
        # copy coordinates file into the project
        path2txt = opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.GCP_COORD_FILE_INIT)[:-4] + ".txt"
        copyfile(gcp_coords_file, path2txt)
        
        success, error = utils.exec_mm3d('mm3d GCPConvert #F={} {}'.format(file_format, path2txt),
                                         display=display)
        if success == 0:
            self.gcp_coord_file = opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.GCP_COORD_FILE_INIT)
            gcp_table = np.loadtxt(path2txt, dtype=str)

            try:
                gcp_name = gcp_table[:, file_format.split('_').index("N")]
                np.savetxt(opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.GCP_NAME_FILE), gcp_name, fmt='%s',
                           newline=os.linesep)
                self.gcp_names = opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.GCP_NAME_FILE)
            except ValueError:  # todo add a coherent except
                print("ERROR prepare_GCP_files(): Check file format and file delimiter. Delimiter is any space")
        else:
            print("ERROR prepare_GCP_files(): Check file format and file delimiter. Delimiter is any space")
            exit(1)
            
    def pick_gcp_ini(self):
        '''
        Function to initialize camera orientation
        
        Pick few GCPs (3 to 5) that MicMac can do a rough estimate of the camera orientation. Then go to pick_gcp_basc() to pick all GCPs of known location
        '''
        if self.gcp_coord_file is None or self.gcp_names is None:
            print("ERROR prepare_gcp_files must be applied first")
        gcp_path = opj(self.project_path, Photo4d.GCP_FOLDER)
        os.chdir(gcp_path)
        print(gcp_path)
        copytree(opj(self.project_path, Photo4d.ORI_FOLDER), opj(gcp_path, 'Ori-Ini'))
        # select the set of image on which to pick GCPs manually
        selected_line = self.sorted_pictures[self.selected_picture_set]
        file_set = "("
        for i in range(len(self.cam_folders)):
            file_set += selected_line[i + 1] + "|"
            in_path = opj(self.cam_folders[i], selected_line[i + 1])
            out_path = opj(gcp_path, selected_line[i + 1])
            copyfile(in_path, out_path)
        file_set = file_set[:-1] + ")"
            
        print('mm3d SaisieAppuisInitQt "{}" Ori-Ini {} {}'.format(file_set, self.GCP_NAME_FILE,
                                                                   self.GCP_PICK_FILE_INI))
        utils.exec_mm3d(
            'mm3d SaisieAppuisInitQt {} Ori-Ini {} {}'.format(file_set, self.GCP_NAME_FILE, self.GCP_PICK_FILE_INI))
        
        command = 'mm3d GCPBascule {} Ori-Ini Bascule-ini {} {}'.format(file_set,
                                                                   self.GCP_COORD_FILE_INIT,
                                                                   self.GCP_PICK_FILE_INI_2D)
        print(command)
        utils.exec_mm3d(command)
        
        try:
            for image in selected_line[1:]:
                os.remove(opj(gcp_path, image))
            for folder in ['Ori-Ini']:
                rmtree(folder)
            #os.remove(Photo4d.GCP_PICK_FILE[:-4] + '-S3D.xml')
            os.chdir(self.project_path)
        except FileNotFoundError:
            pass
        except PermissionError:
            print('WARNING Cannot delete temporary MicMac files due to permission error')
            
    def pick_gcp_basc(self, resolution=5000):
        '''
        Function to pick GCP location on the reference set of images.
        
        Pick all GCPs of known location. Then go the pick_gcp_final() to pick extra points that will also be used as reference buut with coordinates estimated from the camera orientation.
        '''

        if self.gcp_coord_file is None or self.gcp_names is None:
            print("ERROR prepare_gcp_files must be applied first")
        gcp_path = opj(self.project_path, Photo4d.GCP_FOLDER)
        os.chdir(gcp_path)
        print(gcp_path)
        # select the set of image on which to pick GCPs manually
        selected_line = self.sorted_pictures[self.selected_picture_set]
        file_set = "("
        for i in range(len(self.cam_folders)):
            file_set += selected_line[i + 1] + "|"
            in_path = opj(self.cam_folders[i], selected_line[i + 1])
            out_path = opj(gcp_path, selected_line[i + 1])
            copyfile(in_path, out_path)
        file_set = file_set[:-1] + ")"
            
        command='mm3d SaisieAppuisPredicQt "{}" Ori-Bascule-ini {} {}'.format(file_set,
                                                                   self.GCP_COORD_FILE_INIT,
                                                                   self.GCP_PICK_FILE_BASC)
        print(command)        
        utils.exec_mm3d(command)
        
        
        command="mm3d Tapioca All {} {}".format(file_set, resolution)
        print(command)        
        utils.exec_mm3d(command)
        
        command = 'mm3d Campari {} Bascule-ini Bascule GCP=[{},{},{},{}] AllFree=1'.format(file_set, self.GCP_COORD_FILE_INIT, self.GCP_PRECISION, self.GCP_PICK_FILE_BASC_2D, self.GCP_POINTING_PRECISION)
        print(command)
        utils.exec_mm3d(command)
        
        # copy orientation file
        copytree(os.path.join(gcp_path,'Ori-Bascule'), os.path.join(self.project_path,'Ori-Bascule'))
        self.in_ori = os.path.join(self.project_path,'Ori-Bascule')
        
        try:
            for image in selected_line[1:]:
                os.remove(opj(gcp_path, image))
            for folder in ['Ori-Bascule_ini']:
                rmtree(folder)
            #os.remove(Photo4d.GCP_PICK_FILE[:-4] + '-S3D.xml')
            os.chdir(self.project_path)
        except FileNotFoundError:
            pass
        except PermissionError:
            print('WARNING Cannot delete temporary MicMac files due to permission error')
            
    def pick_gcp_final(self):
        '''
        Function to pick additional points that can be set as 'GCPs'. These will get coordinates estimates based on camera orientation, and will be used in other set of images for triangulation.
        This way, we artificailly increase the number of GCPs, and use the selected set of reference images as the absolute reference to which other 3D model will be orientated against. 
        
        Pick as many points as possible that are landmarks across the all set of image.
        '''

        if self.gcp_coord_file is None or self.gcp_names is None:
            print("ERROR prepare_gcp_files must be applied first")
        gcp_path = opj(self.project_path, Photo4d.GCP_FOLDER)
        os.chdir(gcp_path)
        print(gcp_path)
        # select the set of image on which to pick GCPs manually
        selected_line = self.sorted_pictures[self.selected_picture_set]
        file_set = "("
        for i in range(len(self.cam_folders)):
            file_set += selected_line[i + 1] + "|"
            in_path = opj(self.cam_folders[i], selected_line[i + 1])
            out_path = opj(gcp_path, selected_line[i + 1])
            copyfile(in_path, out_path)
        file_set = file_set[:-1] + ")"
            
        command='mm3d SaisieAppuisPredicQt "{}" Ori-Bascule {} {}'.format(file_set,
                                                                   self.GCP_COORD_FILE_INIT,
                                                                   self.GCP_PICK_FILE)
        print(command)        
        utils.exec_mm3d(command)
        self.gcp_coord_file = opj(self.project_path,Photo4d.GCP_FOLDER, Photo4d.GCP_COORD_FILE_FINAL)
        
        try:
            for image in selected_line[1:]:
                os.remove(opj(gcp_path, image))
            for folder in ['Tmp-SaisieAppuis', 'Tmp-MM-Dir','Ori-Bascule','Homol']:
                rmtree(folder)
            #os.remove(Photo4d.GCP_PICK_FILE[:-4] + '-S3D.xml')
            os.chdir(self.project_path)
        except FileNotFoundError:
            pass
        except PermissionError:
            print('WARNING Cannot delete temporary MicMac files due to permission error')
            
    def detect_GCPs(self, kernel_size=(200, 200), display=True, save_df_gcp=True):
        
        xml_file = opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.GCP_PICK_FILE[:-4] + '-S2D.xml')
        self.df_detect_gcp = ds.detect_from_s2d_xml(xml_file, self.cam_folders,
                                                    self.sorted_pictures, kernel_size=kernel_size,
                                                    display_micmac=display)
        if save_df_gcp:
            print("  Saving result to .csv")
            self.df_detect_gcp.to_csv(opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.DF_DETECT_FILE), sep=",")
            
    def extract_GCPs(self, magnitude_max=50, nb_values=5, max_dist=50, kernel_size=(200, 200), method="Median"):
        '''
        Function to sort
        
        Pick all GCPs of known location. Then go the pick_gcp_final() to pick extra points that will also be used as reference buut with coordinates estimated from the camera orientation.
        '''

        if self.df_detect_gcp is None:
            print("ERROR detect_GCPs() must have been run before trying to extract values")
            exit(1)
            
        self.dict_image_gcp, self.df_gcp_abs = ds.extract_values(self.df_detect_gcp, magnitude_max=magnitude_max,
                                                                 nb_values=nb_values, max_dist=max_dist,
                                                                 kernel_size=kernel_size, method=method)
        
        # write xml file for the record
        XML_utils.write_S2D_xmlfile(self.dict_image_gcp,
                                    opj(self.project_path, Photo4d.GCP_FOLDER, Photo4d.GCP_DETECT_FILE ))
        
    def process(self, master_folder=0, clahe=False, tileGridSize_clahe=8, resol=-1,
                re_estimate=True,
                DefCor=0.0, delete_temp=True, display=True):
        if self.sorted_pictures is None:
            print("ERROR You must apply sort_pictures() before doing anything else")
            exit(1)
        print(self.in_ori, "LAAAAA")
        
        proc.process_from_array(self.cam_folders, self.sorted_pictures, opj(self.project_path, Photo4d.RESULT_FOLDER),
                                inori=self.in_ori,
                                gcp=self.gcp_coord_file, gcp_S2D=self.dict_image_gcp, clahe=clahe,
                                tileGridSize_clahe=tileGridSize_clahe, resol=resol, distortion_model=self.distortion_model,
                                re_estimate=re_estimate, master_folder=master_folder, masq2D=self.masks, DefCor=DefCor,
                                shift=self.shift, delete_temp=delete_temp,
                                display_micmac=display, cond=self.cond, 
                                GNSS_PRECISION=self.GCP_PRECISION, GCP_POINTING_PRECISION=self.GCP_POINTING_PRECISION)
        
    def set_selected_set(self, img_or_index: Union[int, str]):
        if self.sorted_pictures is None:
            print("ERROR You must apply sort_pictures before trying to chose a set")
            exit(1)
        else:
            if type(img_or_index) == int:
                self.selected_picture_set = img_or_index
                print(
                    "\n The current selected set is now {}".format(self.sorted_pictures[self.selected_picture_set][1:]))
            elif type(img_or_index) == str:
                found, i = False, 0
                while (not found) and (i < len(self.sorted_pictures)):
                    if img_or_index in self.sorted_pictures[i]:
                        found = True
                        self.selected_picture_set = i
                        print("\n The current selected set is now {}".format(
                            self.sorted_pictures[self.selected_picture_set][1:]))
                    i += 1
                if not found:
                    print('image {} not in sorted_pictures'.format(img_or_index))
                    
    def add_cond(self, condition):
        if type(condition) != function:
            print("WARNING input condition must be a function, which return a boolean and take a date as parameter")
        else:
            print("Condition applied")
    # function to only execute program on a subseltection of images
        
    def graph_detected_gcps(self, nbInputGCPs=0):
        max_dist=50
        magnitude_max=50
        kernel_size=(200, 200)
        df=self.df_detect_gcp
        
        df['Xshift'] = df.Xgcp_ini + df.Xdetect - df.Xini
        df['Yshift'] = df.Ygcp_ini + df.Ydetect - df.Yini
        
        # compute vector module
        df['magnitude'] = np.sqrt((df.Xini - df.Xdetect) ** 2 + (df.Yini - df.Ydetect) ** 2)
        
        # compute vector direction
        df['direction'] = np.arctan2((df.Xini - df.Xdetect), (df.Yini - df.Ydetect)) * 180 / np.pi + 180
        
        # compute from gcp and tie point in the initial image (gcp is in the center of the extracts)
        pos_center = kernel_size[0] / 2, kernel_size[1] / 2
        df['dist'] = np.sqrt((df.Xini - pos_center[0]) ** 2 + (df.Yini - pos_center[1]) ** 2)
        
        # filter outliers having a incoherent magnitude
        df_filtered = df.loc[df.magnitude <= magnitude_max]
        
        
        plt.figure(figsize=(20, 10))
        plt.subplot(121)
        
        GCP_names_all=[]
        Img_names_all=[]
        for i in range(0,3):
            df_filtered_middle = df_filtered.loc[df_filtered.folder_index == i]
            nbGCPs=[]
            GCP_names_onestack=[]
            Img_names_onestack=[]
            for date, group in df_filtered_middle.groupby(['date']):
                GCP_names_oneim=[]
                Img_names_onestack.append(group.Image[:1])
                nbGCP=0
                for gcp, group_gcp in group.groupby(['GCP']):
                    group_gcp_filtered = group_gcp.loc[group_gcp.dist <= max_dist]
                    if(len(group_gcp_filtered)>0):
                        GCP_names_oneim.append(gcp)
                        nbGCP=nbGCP+1
                        #gcps=gcps + " " + gcp
                nbGCPs.append(nbGCP)
                #print('{}'.format(gcps))
                #print('NbGCP : {}'.format(nbGCP))
                #print('')
                GCP_names_onestack.append(GCP_names_oneim)
            GCP_names_all.append(GCP_names_onestack)
            Img_names_all.append(Img_names_onestack)
            plt.plot(nbGCPs)
        
        plt.title('GCP found in each image for each 3 stacks')
        plt.grid(True)
        plt.xlabel('Image group index')
        plt.ylabel('Number of points')
        plt.ylim(0,nbInputGCPs)
        
        GCP_names_common=[]
        GCP_nb_common=[]
        print("--------------------------------------")
        for i in range(0,len(self.sorted_pictures)-1):
            GCP_names_common_oneSet = set(set(GCP_names_all[2][i]).intersection(GCP_names_all[1][i])).intersection(GCP_names_all[0][i])
            GCP_names_common.append(GCP_names_common_oneSet)
            print("{} {} {}".format(Img_names_all[0][i].values[0],Img_names_all[1][i].values[0],Img_names_all[2][i].values[0]))
            print("Number of points : {}".format(len(GCP_names_common_oneSet)))
            print(GCP_names_common_oneSet)
            print("--------------------------------------")
            GCP_nb_common.append(len(GCP_names_common_oneSet))
        plt.subplot(122)    
        plt.plot(GCP_nb_common)
        plt.title('GCP found in all images of each timestamp')
        plt.grid(True)
        plt.xlabel('Image group index')
        plt.ylabel('Number of points')
        plt.ylim(0,nbInputGCPs)
        
if __name__ == "__main__":
    myproj = Photo4d(project_path=r"I:\icemass-users\lucg\Finse\Photo4D\2018-1pm")
    myproj = Photo4d(project_path=r"~/icemassME/Finse/Photo4D/2018-1pm")
   # myproj.sort_picture()
    #myproj.check_picture()
    #myproj.set_selected_set("DSC03111.JPG")

    #myproj.initial_orientation()
    #myproj.create_mask()
    #myproj.prepare_gcp_files(r"I:\icemass-users\lucg\Finse\Photo4D\2018-1pm\GCPs_coordinates_GNSS_Only.txt",file_format="N_X_Y_Z")
    #myproj.pick_gcp_ini()
    #myproj.pick_gcp_basc()
    #myproj.pick_gcp_final()
    #myproj.detect_GCPs()
    #myproj.extract_GCPs()
    #myproj.graph_detected_gcps(50)
    #myproj.process(resol=4000)
    # Part 1, sort and flag good picture set
    # myproj.sort_picture()
    # myproj.check_picture()

    # Part 2: Manual
    # choose either:
    # 		- Calibration, calibration of camera (TODO)
    # 		- Orientation, estimate camera position

    # myproj.selected_picture_set = -1  # by default get the last set of image. Change to the correct index where there are good quality images
    # myproj.oriententation_intitial()

    # myproj.selected_picture_set = -1
    # myproj.create_mask()

    # Part 3: Derive GCPS in image stack

    # myproj.prepare_GCP_files("GCPs.txt")  # format the text file to

    # myproj.selected_picture_set = -1
    # myproj.pick_GCP()

    # myproj.detect_GCPs()

    # point GCP (manual)
    # estimate on stack

    # Part 4: Process images to point cloud
    # myproj.process()
